import os

import urllib3

from src.cli.helper import exist_or_create
from src.local.providers.types import K8sProviderType

# disable warnings
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# cli
CLI_CONFIG_FILE = os.path.expanduser("~/.unikube/config_dev")
exist_or_create(CLI_CONFIG_FILE)

CLI_KUBECONFIG_DIRECTORY = os.path.expanduser("~/.unikube/")
CLI_TABLEFMT = "psql"

CLI_LOG_LEVEL = "DEBUG"  # DEBUG, INFO, WARNING, ERROR/SUCCESS
CLI_ALWAYS_SHOW_CONTEXT = False

# authentication
AUTH_DEFAULT_HOST = "https://login.unikube.io"  # "http://keycloak.127.0.0.1.nip.io:8085"

# unikube
UNIKUBE_FILE = "unikube.yaml"
UNIKUBE_DEFAULT_PROVIDER_TYPE = K8sProviderType.k3d

# token
TOKEN_REALM = "unikube"
TOKEN_ALGORITHM = "RS256"
TOKEN_PUBLIC_KEY = f"/auth/realms/{TOKEN_REALM}"
TOKEN_LOGIN_PATH = f"/auth/realms/{TOKEN_REALM}/protocol/openid-connect/token"
TOKEN_VERIFY_PATH = f"/auth/realms/{TOKEN_REALM}/protocol/openid-connect/userinfo"
TOKEN_REFRESH_PATH = f"/auth/realms/{TOKEN_REALM}/protocol/openid-connect/token"
TOKEN_TIMEOUT = 30
TOKEN_AUDIENCE = "gateway"
TOKEN_RPT_AUDIENCE = "gateway"
KC_CLIENT_ID = "cli"

# GraphQL
GRAPHQL_URL = "https://api.unikube.io/graphql/"  # "http://gateway.unikube.127.0.0.1.nip.io:8085/graphql/"
GRAPHQL_TIMEOUT = 30

# manifest
MANIFEST_DEFAULT_HOST = "https://api.unikube.io/manifests/"

# local system: dependencies + versions + settings
DOCKER_TEST_IMAGE = "busybox"
DOCKER_CLI_MIN_VERSION = "15.0.1"
DOCKER_WEBSITE = "https://docs.docker.com/install/"

K3S_CLI_MIN_VERSION = "1.17.1"

K3D_CLI_MIN_VERSION = "3.0.0"
K3D_WEBSITE = "https://github.com/rancher/k3d"
K3D_CLUSTER_PREFIX = "unikube-"
K3D_DEFAULT_INGRESS_PORT = 80
K3D_DEFAULT_WORKERS = 1

TELEPRESENCE_CLI_MIN_VERSION = "0.101"
TELEPRESENCE_TAG_PREFIX = "telepresence:dev"
TELEPRESENCE_DOCKER_IMAGE_FORMAT = "{project}-{deck}-{name}-" + TELEPRESENCE_TAG_PREFIX

HOMEBREW_CLI_VERSION = "2.1.0"
HOMEBREW_WEBSITE = "https://brew.sh/"

KUBECTL_MIN_CLI_VERSION = "1.18.0"



from typing import List

from pydantic import BaseModel
from tinydb import Query, TinyDB

from src import settings
from src.storage.types import TinyDatabaseData


class TinyDatabase:
    def __init__(
        self,
        table_name="database",
    ):
        self.table_name = table_name

        self.db = TinyDB(settings.CLI_CONFIG_FILE)
        self.table = self.db.table(self.table_name)

    def select(self, id: str) -> TinyDatabaseData:
        document = self.table.get(Query().id == id)
        return document

    def insert(self, data: BaseModel) -> int:
        doc_id = self.table.insert(data.dict())
        return doc_id

    def update(self, id: str, data: BaseModel) -> List[int]:
        doc_id_list = self.table.upsert(data.dict(), Query().id == id)
        return doc_id_list

    def delete(self, id: str) -> bool:
        doc_id = self.table.remove(Query().id == id)
        if not doc_id:
            return False
        return True

    def drop(self):
        self.db.purge_table(self.table_name)



from urllib.parse import urljoin

import click_spinner
import requests
from requests import Session

from src import settings


def get_requests_session(access_token) -> Session:
    session = requests.Session()
    session.headers.update({"Content-type": "application/json", "Authorization": "Bearer " + str(access_token)})
    return session


def download_specs(access_token: str, environment_id: str):
    session = get_requests_session(access_token=access_token)

    manifest_url = urljoin(settings.MANIFEST_DEFAULT_HOST, environment_id)
    with click_spinner.spinner():
        r = session.get(manifest_url)
    r.raise_for_status()

    manifest = r.json()
    return manifest



import os

import click
import click_spinner

from src import settings
from src.cli import console
from src.graphql import GraphQL
from src.local.system import Docker, KubeAPI, KubeCtl, Telepresence
from src.settings import UNIKUBE_FILE
from src.unikubefile.selector import unikube_file_selector


def get_required_information(ctx, project_title: str, deck_title: str):
    ## project_id
    cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)
    cluster_title_list = [item.name for item in cluster_list]

    # argument
    if not project_title:
        # argument from console
        project_title = console.list(
            message="Please select a project",
            message_no_choices="No cluster is running.",
            choices=cluster_title_list,
        )
        if project_title is None:
            console.exit_generic_error()

    # check if project is in local storage
    if project_title not in cluster_title_list:
        console.error("The project cluster could not be found.", _exit=True)

    # get project_id
    project_id = None
    cluster = None
    for cluster_data in cluster_list:
        if cluster_data.name == project_title:
            cluster = ctx.cluster_manager.select(
                cluster_data=cluster_data,
            )
            project_id = cluster.id
            break

    if not project_id:
        console.error("The project id could not be determined.", _exit=True)

    ## deck_id
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                allDecks(projectId: $id) {
                    results {
                        id
                        title
                        namespace
                    }
                }
            }
            """,
            query_variables={
                "id": project_id,
            },
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    deck_list = data["allDecks"]["results"]
    deck_title_list = [item["title"] for item in deck_list]

    # argument
    if not deck_title:
        # argument from console
        deck_title = console.list(
            message="Please select a deck",
            message_no_choices="No deck found.",
            choices=deck_title_list,
        )
        if deck_title is None:
            console.exit_generic_error()

    # check if deck exists
    if deck_title not in deck_title_list:
        console.error("The deck could not be found.", _exit=True)

    # get deck
    deck = None
    for deck in deck_list:
        if deck["title"] == deck_title:
            break

    return project_id, project_title, deck


@click.command()
def list(**kwargs):
    raise NotImplementedError


@click.command()
def info(**kwargs):
    raise NotImplementedError


@click.command()
def use(**kwargs):
    raise NotImplementedError


@click.command()
@click.argument("project_title", required=False)
@click.argument("deck_title", required=False)
@click.argument("pod_title", required=False)
@click.pass_obj
def shell(ctx, project_title, deck_title, pod_title, **kwargs):
    """Drop into an interactive shell."""

    ctx.auth.check()

    project_id, project_title, deck = get_required_information(ctx, project_title, deck_title)

    ## shell
    # check if cluster is ready
    cluster_data = ctx.cluster_manager.get(id=project_id)
    print(project_id)
    cluster = ctx.cluster_manager.select(cluster_data=cluster_data)
    if not cluster:
        console.error("The project cluster does not exist.")
        return None

    provider_data = cluster.storage.get()

    # shell
    k8s = KubeAPI(provider_data, deck)

    if not pod_title:
        pod_list_choices = [pod.metadata.name for pod in k8s.get_pods().items]
        pod_title = console.list(
            message="Please select a pod",
            choices=pod_list_choices,
        )

    if not pod_title:
        console.error("No pods available.")
        return None

    # get the data of the selected pod
    data = k8s.get_pod(pod_title)

    # 1. check if this pod is of a switched deployment (in case of an active Telepresence)
    if data.metadata.labels.get("telepresence"):
        # the corresponding deployment by getting rid of the telepresence suffix
        deployment = "-".join(data.metadata.name.split("-")[0:-1])

        # the container name generated in "app switch" for that pod
        container_name = settings.TELEPRESENCE_DOCKER_IMAGE_FORMAT.format(
            project=project_title, deck=deck["title"], name=deployment
        ).replace(":", "")

        if Docker().check_running(container_name):
            # 2. Connect to that container
            # 2.a connect using Docker
            Docker().exec(container_name, "/bin/sh", interactive=True)
        else:
            console.error(
                "This is a Telepresence Pod with no corresponding Docker container "
                "running in order to connect (inconsistent state?)"
            )

    else:
        # 2.b connect using kubernetes
        KubeCtl(provider_data).exec_pod(pod_title, deck["namespace"], "/bin/sh", interactive=True)


@click.command()
@click.argument("project_title", required=False)
@click.argument("deck_title", required=False)
@click.option("--deployment", help="Specify the deployment if not set in the Unikubefile")
@click.option("--image", help="Specify the Docker image from your local registry")
@click.option("--unikubefile", help="Specify the path to the Unikubefile", type=str)
@click.pass_obj
def switch(ctx, project_title, deck_title, deployment, image, unikubefile, **kwargs):
    """Switch a running deployment with a local Docker image"""

    ctx.auth.check()

    project_id, project_title, deck = get_required_information(ctx, project_title, deck_title)

    ## switch
    # check if cluster is ready
    cluster_data = ctx.cluster_manager.get(id=project_id)
    cluster = ctx.cluster_manager.select(cluster_data=cluster_data)
    if not cluster:
        console.error("The project cluster does not exist.")
        return None

    # unikube file input
    if unikubefile:
        path_unikube_file = unikubefile
    else:
        path_unikube_file = os.path.join(os.getcwd(), UNIKUBE_FILE)

    unikube_file = unikube_file_selector.get(path_unikube_file=path_unikube_file)

    # 2: Get a deployment
    # 2.1.a Check the deployment identifier
    if not deployment:
        # 1.1.b check the unikubefile
        deployment = unikube_file.get_deployment()
        if not deployment:
            console.error("Please specify the deployment either using the '--deployment' option or in the Unikubefile")

    # 2.2 Fetch available "deployment:", deployments
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                deck(id: $id) {
                    deployments(level: "local") {
                        id
                        title
                        description
                        ports
                        isSwitchable
                    }
                }
            }
            """,
            query_variables={
                "id": deck["id"],
            },
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    target_deployment = None

    for _deployment in data["deck"]["deployments"]:
        if _deployment["title"] == deployment:
            target_deployment = _deployment

    # 2.3 Check and select deployment data
    if target_deployment is None:
        console.error(
            f"The deployment '{deployment}' you specified could not be found.",
            _exit=True,
        )

    ports = target_deployment["ports"].split(",")
    deployment = target_deployment["title"]
    namespace = deck["namespace"]

    # 3: Build an new Docker image
    # 3.1 Grab the docker file
    context, dockerfile, target = unikube_file.get_docker_build()
    console.debug(f"{context}, {dockerfile}, {target}")
    console.info(f"Building a Docker image for {dockerfile} with context {context}")

    # 3.2 Set an image name
    image_name = settings.TELEPRESENCE_DOCKER_IMAGE_FORMAT.format(
        project=project_title.replace(" ", "").lower(), deck=deck["title"], name=deployment
    )

    # 3.3 Build image
    docker = Docker()
    with click_spinner.spinner():
        status, msg = docker.build(image_name, context, dockerfile, target)
    if not status:
        console.debug(msg)
        console.error("Failed to build Docker image.", _exit=True)

    console.info(f"Docker image successfully built: {image_name}")

    # 4. Start the Telepresence session
    # 4.1 See if there are volume mounts
    mounts = unikube_file.get_mounts()
    console.debug(f"Volumes requested: {mounts}")

    # 4.2 See if there special env variables
    envs = unikube_file.get_environment()
    console.debug(f"Envs requested: {envs}")

    # 4.3 See if there is a run command to be executed
    command = unikube_file.get_command(port=ports[0])
    console.debug(f"Run command: {command}")

    console.info("Starting your container, this may takes a while to become effective")
    provider_data = cluster.storage.get()
    Telepresence(provider_data, debug_output=True).swap(deployment, image_name, command, namespace, envs, mounts)
    # if something went wrong with Telepresence
    if docker.check_running(image_name):
        docker.kill(name=image_name)


@click.command()
def pulldb(**kwargs):
    raise NotImplementedError


@click.command()
@click.argument("project_title", required=False)
@click.argument("deck_title", required=False)
@click.argument("pod_title", required=False)
@click.argument("pod_title", required=False)
@click.option("--watch", "-w", is_flag=True, default=False, help="Watch logs.")
@click.pass_obj
def logs(ctx, project_title, deck_title, pod_title, watch, **kwargs):
    """Display the container's logs"""

    ctx.auth.check()

    project_id, project_title, deck = get_required_information(ctx, project_title, deck_title)

    ## logs
    # check if cluster is ready
    cluster_data = ctx.cluster_manager.get(id=project_id)
    cluster = ctx.cluster_manager.select(cluster_data=cluster_data)
    if not cluster:
        console.error("The project cluster does not exist.")
        return None

    provider_data = cluster.storage.get()

    # log
    k8s = KubeAPI(provider_data, deck)

    if not pod_title:
        pod_list_choices = [pod.metadata.name for pod in k8s.get_pods().items]
        pod_title = console.list(
            message="Please select a pod",
            choices=pod_list_choices,
        )

    if not pod_title:
        console.error("No pods available.")
        return None

    logs = k8s.get_logs(pod_title, watch)

    # output
    click.echo(logs)


@click.command()
def expose(**kwargs):
    raise NotImplementedError


@click.command()
def env(**kwargs):
    raise NotImplementedError


@click.command()
def request_env(**kwargs):
    raise NotImplementedError


@click.command()
def exec(**kwargs):
    raise NotImplementedError



import os
from abc import ABC, abstractmethod
from typing import List, Union
from uuid import UUID

from src import settings
from src.cli import console
from src.context.types import ContextData
from src.graphql import GraphQL
from src.storage.user import LocalStorageUser, get_local_storage_user
from src.unikubefile.selector import unikube_file_selector
from src.unikubefile.unikube_file import UnikubeFile


class IContext(ABC):
    @abstractmethod
    def get(self, **kwargs) -> ContextData:
        raise NotImplementedError


class ClickOptionContext(IContext):
    def __init__(self, click_options):
        self.click_options = click_options

    def get(self, **kwargs) -> ContextData:
        # uuid validation
        def is_valid_uuid4(uuid):
            try:
                _ = UUID(uuid, version=4)
                return True
            except Exception:
                return False

        # organization
        organization = self.click_options.get("organization", None)
        if organization:
            if is_valid_uuid4(organization):
                organization_id = organization
            else:
                raise NotImplementedError
        else:
            organization_id = None

        # project
        project = self.click_options.get("project", None)
        if project:
            if is_valid_uuid4(organization):
                project_id = project
            else:
                raise NotImplementedError
        else:
            project_id = None

        # deck
        deck = self.click_options.get("deck", None)
        if deck:
            if is_valid_uuid4(deck):
                deck_id = deck
            else:
                raise NotImplementedError
        else:
            deck_id = None

        return ContextData(
            organization_id=organization_id,
            project_id=project_id,
            deck_id=deck_id,
        )


class UnikubeFileContext(IContext):
    def __init__(self, unikube_file: Union[UnikubeFile, None]):
        self.unikube_file = unikube_file

    def get(self, **kwargs) -> ContextData:
        # check if unikube file was loaded
        if not self.unikube_file:
            return ContextData()

        return self.unikube_file.get_context()


class LocalContext(IContext):
    def __init__(self, local_storage_user: Union[LocalStorageUser, None]):
        self.local_storage_user = local_storage_user

    def get(self, **kwargs) -> ContextData:
        if not self.local_storage_user:
            return ContextData()

        user_data = self.local_storage_user.get()
        return user_data.context


class ImplizitContext(IContext):
    """Implicit context, e.g.: user has only one organization"""

    def get(self, **kwargs) -> ContextData:
        return ContextData()


class ContextLogic:
    def __init__(self, context_order: List[IContext]):
        self.context_order = context_order

    def get(self) -> ContextData:
        context = ContextData()

        for context_object in self.context_order:
            # get context variables from current context
            context_current = context_object.get(current_context=context)

            # update context
            context_dict = context.dict()
            for key, value in context_current.dict().items():
                if context_dict[key] is None:
                    context_dict[key] = value
            context = ContextData(**context_dict)

            # check if all context variables have already been set
            if None not in context.dict().values():
                break

        return context


class Context:
    def __init__(self, auth):
        self._auth = auth

    def get(self, **kwargs) -> ContextData:
        local_storage_user = get_local_storage_user()

        context_logic = ContextLogic(
            [
                ClickOptionContext(
                    click_options={key: kwargs[key] for key in ("organization", "project", "deck") if key in kwargs}
                ),
                UnikubeFileContext(
                    unikube_file=unikube_file_selector.get(
                        path_unikube_file=os.path.join(
                            os.getcwd(),
                            "unikube.yaml",
                        )
                    )
                ),
                LocalContext(local_storage_user=local_storage_user),
                ImplizitContext(),
            ]
        )
        context = context_logic.get()

        # show context
        if settings.CLI_ALWAYS_SHOW_CONTEXT:
            console.info(f"context: {context}")

        return context

    def __graph_ql(self, query: str, query_variables: dict) -> Union[dict, None]:
        # GraphQL
        try:
            graph_ql = GraphQL(authentication=self._auth)
            result = graph_ql.query(
                query,
                query_variables=query_variables,
            )
            key = next(iter(result))
            data = result[key]
        except Exception as e:
            data = None
            console.debug(e)
            console.exit_generic_error()

        return data

    def get_organization(self) -> Union[dict, None]:
        return self.__graph_ql(
            """
            query($id: UUID!) {
                organization(id: $id) {
                    title
                    id
                }
            }
            """,
            query_variables={
                "id": self.get().organization_id,
            },
        )

    def get_project(self) -> Union[dict, None]:
        return self.__graph_ql(
            """
            query($id: UUID) {
                project(id: $id) {
                    title
                    id
                }
            }
            """,
            query_variables={
                "id": self.get().project_id,
            },
        )

    def get_deck(self) -> Union[dict, None]:
        return self.__graph_ql(
            """
            query($id: UUID) {
                deck(id: $id) {
                    title
                    id
                }
            }
            """,
            query_variables={
                "id": self.get().deck_id,
            },
        )



class ClickContext(object):
    def __init__(self):
        from src.authentication.authentication import get_authentication
        from src.context.context import Context
        from src.local.providers.manager import K8sClusterManager
        from src.storage.general import LocalStorageGeneral

        self.auth = get_authentication()
        self.storage_general = LocalStorageGeneral()
        self.context = Context(auth=self.auth)
        self.cluster_manager = K8sClusterManager()



from abc import ABC, abstractmethod

from src.storage.tinydb import TinyDatabase
from src.storage.types import TinyDatabaseData


class ILocalStorage(ABC):
    @abstractmethod
    def get(self, id: str):
        raise NotImplementedError

    @abstractmethod
    def set(self, id: str, data: TinyDatabaseData):
        raise NotImplementedError

    @abstractmethod
    def delete(self, id: str):
        raise NotImplementedError


class LocalStorage(ILocalStorage):
    table_name = "local"
    pydantic_class = TinyDatabaseData

    def __init__(self) -> None:
        # database / storage
        self.database = TinyDatabase(table_name=self.table_name)

    def get(self, id: str, **kwargs) -> TinyDatabaseData:
        try:
            document = self.database.select(id=id)
            return self.pydantic_class(**dict(document))
        except Exception:
            return self.pydantic_class(id=id, **kwargs)

    def set(self, id: str, data: TinyDatabaseData) -> None:
        self.database.update(id=id, data=data)

    def delete(self, id: str) -> None:
        self.database.delete(id=id)



import pytest
from click.testing import CliRunner

from src.cli import auth
from unikube import ClickContext


def test_login():
    runner = CliRunner()
    result = runner.invoke(
        auth.login,
        ["--email", "test@test.de", "--password", "secure"],
        obj=ClickContext(),
    )
    assert result.output == "[ERROR] Login failed. Please check your e-mail or re-enter your password.\n"
    assert result.exit_code == 0


def test_logout():
    runner = CliRunner()
    result = runner.invoke(
        auth.logout,
        obj=ClickContext(),
    )
    assert result.output == "[INFO] Logout completed.\n"
    assert result.exit_code == 0



# -*- coding: utf-8 -*-
import os
import platform
import re
import subprocess
from time import sleep
from typing import Dict, List, Optional, Tuple

import click
import semantic_version

import src.cli.console as console
from src import settings


class LocalDependency(object):
    """
    This is a local development dependency unikube
    """

    cmd = None
    verbose_name = None
    required = None
    required_version = None

    def probe(self, silent) -> Tuple[bool, str]:
        console.info(self.check_message(), silent=silent, nl=False)

        if not self.required_version:
            try:
                kwargs = {
                    "stdout": open(os.devnull, "w"),
                    "close_fds": True,
                    "stderr": subprocess.STDOUT,
                }
                subprocess.check_call(self.cmd, **kwargs)
                self._click_echo(" Ok", silent=silent, fg="green")
                return True, ""
            except (subprocess.CalledProcessError, FileNotFoundError):
                self._click_echo(" Error", silent=silent, fg="red")
                return False, self.notfound_message()
        else:
            try:
                pout = subprocess.run(self.cmd, capture_output=True, encoding="UTF-8")
            except (subprocess.CalledProcessError, FileNotFoundError):
                self._click_echo(" Error", silent=silent, fg="red")
                return False, self.notfound_message()
            else:
                cmd_version = self.prepare_version_string(pout.stdout)
                version, sufficient = self.check_version(cmd_version)
                if version and sufficient:
                    self._click_echo(f" Ok (Version {version})", silent=silent, fg="green")
                    return True, ""
                elif version:
                    self._click_echo(f" Error (Version {version})", silent=silent, fg="red")
                    return False, self.oldversion_message()
                else:
                    self._click_echo(" Error", silent=silent, fg="red")
                    return False, self.notfound_message()

    def _click_echo(self, msg, silent=False, fg=""):
        if not silent:
            click.secho(msg, fg=fg)

    def install(self) -> int:
        return_code = 0
        for idx, step in enumerate(self.installation_steps):
            console.info(f"Running installation step #{idx+1}: {step}")
            try:
                process = subprocess.run(step, shell=True)
                return_code += process.returncode
            except Exception as cpr:
                console.error(f"An error occured during the installation of {self.verbose_name}:")
                console.error(str(cpr))
                return 1
        return return_code

    def notfound_message(self) -> str:
        return f"Please make sure that '{self.verbose_name}' is correctly installed on your computer."

    def oldversion_message(self) -> str:
        return f"The version of '{self.verbose_name}' on your computer is too old."

    def check_message(self) -> str:
        return f"Checking {self.verbose_name} "

    def check_version(self, cmd_version) -> Tuple[str, bool]:
        self.installed_version = semantic_version.Version(cmd_version)
        required_version = semantic_version.Version(self.required_version)
        if required_version <= self.installed_version:
            return str(self.installed_version), True
        else:
            return str(self.installed_version), False

    def prepare_version_string(self, cmd_version) -> str:
        return cmd_version.strip()


class Kubectl(LocalDependency):
    cmd = ("kubectl", "version", "--client")
    verbose_name = "Kubectl"
    required_version = settings.KUBECTL_MIN_CLI_VERSION

    @property
    def installation_steps(self):
        # Check for OS
        if platform.system() == "Darwin":
            return [
                "curl -LO https://storage.googleapis.com/kubernetes-release/release/$"
                "(curl -s https://storage.googleapis.com/kubernetes-release/release/stable.txt)"
                "/bin/darwin/amd64/kubectl",
                "chmod +x ./kubectl",
                "sudo mv ./kubectl /usr/local/bin/kubectl",
            ]

        return [
            "curl -LO https://storage.googleapis.com/kubernetes-release/release/$"
            "(curl -s https://storage.googleapis.com/kubernetes-release/release/stable.txt)"
            "/bin/linux/amd64/kubectl",
            "chmod +x ./kubectl",
            "sudo mv ./kubectl /usr/local/bin/kubectl",
        ]

    def prepare_version_string(self, cmd_version) -> str:
        # version.Info{Major:"1", Minor:"19", GitVersion:"v1.19.0", GitCommit:"e19964183377d0ec2052d1f1fa930c4d7575bd50"
        # , GitTreeState:"clean", BuildDate:"2020-08-26T14:30:33Z", GoVersion:"go1.15",
        # Compiler:"gc", Platform:"linux/amd64"}

        match = re.search(r'GitVersion:"(.+?)"', cmd_version).group(1)
        return match.strip("v")


class K3s(LocalDependency):
    cmd = ("k3s", "--version")
    verbose_name = "K3S"
    required_version = settings.K3S_CLI_MIN_VERSION

    def prepare_version_string(self, cmd_version) -> str:
        # k3s version vx.yy.z+k3s1 (12345678)
        return cmd_version.replace("k3s version ", "").replace("v", "").split(" ")[0]

    @property
    def installation_steps(self):
        # Check for OS
        if platform.system() == "Darwin":
            return ["curl -sfL https://get.k3s.io | sh -"]
        return ["curl -sfL https://get.k3s.io | sudo bash"]


class K3d(LocalDependency):
    cmd = ("k3d", "--version")
    verbose_name = "k3d"
    website = settings.K3D_WEBSITE
    required_version = settings.K3D_CLI_MIN_VERSION
    installation_steps = ["curl -s https://raw.githubusercontent.com/rancher/k3d/main/install.sh | sudo bash"]

    def prepare_version_string(self, cmd_version) -> str:
        # k3d version vx.y.z
        # k3s version v1.18.6-k3s1 (default)
        return cmd_version.split("\n")[0].replace("k3d version ", "").replace("v", "").split(" ")[0]


class DockerEngine(LocalDependency):
    cmd = ("docker", "run", "--rm", settings.DOCKER_TEST_IMAGE, "true")
    verbose_name = "Docker Engine"


class Docker(LocalDependency):
    cmd = ("docker", "--version")
    verbose_name = "Docker"
    required_version = settings.DOCKER_CLI_MIN_VERSION
    website = settings.DOCKER_WEBSITE

    def prepare_version_string(self, cmd_version) -> str:
        # Docker version xx.yy.z, build 1234567
        cmd_version = cmd_version.replace("Docker version ", "").split(",")[0]
        # Docker uses a slightly incorrect semantic versioning:
        # e.g. Docker version 19.03.8, build afacb8b7f0 -> ValueError: Invalid leading zero in minor: '19.03.8'
        cmd_version = ".".join(str(int(bit)) for bit in cmd_version.split("."))
        return cmd_version


class Telepresence(LocalDependency):
    cmd = ("telepresence", "--version")
    verbose_name = "Telepresence"
    required_version = settings.TELEPRESENCE_CLI_MIN_VERSION

    @property
    def installation_steps(self):
        if platform.system() == "Darwin":
            return [
                "brew cask install osxfuse",
                "brew install datawire/blackbird/telepresence",
            ]
        return [
            "curl -s https://packagecloud.io/install/repositories/datawireio/telepresence/script.deb.sh | sudo bash",
            "sudo apt install -y --no-install-recommends telepresence",
        ]

    def check_version(self, cmd_version) -> Tuple[str, bool]:
        # Telepresence does not support SemVer
        self.installed_version = re.sub(r"[^0-9\.]", "", cmd_version)
        self.required_version = re.sub(r"[^0-9\.]", "", self.required_version)
        required_major, required_minor = map(int, self.required_version.split(".")[0:2])
        installed_major, installed_minor = map(int, self.installed_version.split(".")[0:2])
        # simple check for major and minor version
        if installed_major > required_major:
            return str(self.installed_version), True
        elif installed_major >= required_major and installed_minor >= required_minor:
            return str(self.installed_version), True
        else:
            return str(self.installed_version), False


class Homebrew(LocalDependency):
    cmd = ("brew", "--version")
    verbose_name = "Homebrew"
    required_version = settings.HOMEBREW_CLI_VERSION
    installation_steps = [
        "/bin/bash -c $(curl -fsSL " "https://raw.githubusercontent.com/Homebrew/install/master/install.sh)"
    ]
    website = settings.HOMEBREW_WEBSITE

    def prepare_version_string(self, cmd_version) -> str:
        return cmd_version.split("\n")[0].replace("Homebrew ", "")


ALL_DEPENDENCIES = [
    Docker,
    DockerEngine,
    Kubectl,
    K3d,
    Telepresence,
]

if platform.system() == "Darwin":
    ALL_DEPENDENCIES = [Homebrew] + ALL_DEPENDENCIES


def probe_dependencies(silent=False) -> List[Dict[str, str]]:
    """Generates a report of the required software and versions"""
    results = []
    for klass in ALL_DEPENDENCIES:
        check = klass()
        success, msg = check.probe(silent=silent)
        if check.required_version:
            required_version = check.required_version
            if success:
                installed_version = check.installed_version
            else:
                installed_version = ""
        else:
            required_version = ""
            installed_version = ""
        results.append(
            {
                "name": check.verbose_name,
                "success": success,
                "msg": msg,
                "required_version": required_version,
                "installed_version": installed_version,
            }
        )
    return results


def install_dependency(name, silent=False) -> Optional[int]:
    try:
        klass = next(
            filter(
                lambda klass: klass.verbose_name.lower() == name.lower(),
                ALL_DEPENDENCIES,
            )
        )
    except StopIteration:
        console.error(f"The dependency name '{name}' is not valid. No action taken.")
        return None
    else:
        if hasattr(klass, "installation_steps"):
            console.info(f"Now running installation setup for {klass.verbose_name}")
            rcode = klass().install()
            if rcode == 0:
                console.success(f"The installation of {klass.verbose_name} was successful.")
            else:
                console.error(f"The installation of {klass.verbose_name} was not successful.")
            return rcode
        else:
            console.error(
                f"The unikube.tech CLI does currently not support the installation of {klass.verbose_name}. "
                f"{'Please find instructions here: ' + klass.website if hasattr(klass, 'website') else ''}"
            )
            return None



from src.local.providers.abstract_provider import AbstractK8SProviderStorage
from src.local.providers.k3d.types import K3dData
from src.storage.local_storage import LocalStorage


class K3dLocalStorage(LocalStorage):
    table_name = "k3d"
    pydantic_class = K3dData


class K3dStorage(AbstractK8SProviderStorage):
    def __init__(self, id: str) -> None:
        super().__init__(
            id=id,
            storage=K3dLocalStorage(),
        )



import os

from setuptools import find_packages, setup

base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, "VERSION")) as f:
    VERSION = f.read()


with open(os.path.join(base_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

DESCRIPTION = "This is the unikube.io command line interface"


setup(
    name="unikube",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=VERSION,
    py_modules=["unikube"],
    install_requires=[
        "click==7.1.2",
        "click-spinner==0.1.10",
        "colorama~=0.4.4",
        "inquirer~=2.7.0",
        "tabulate~=0.8.7",
        "pydantic~=1.7.3",
        "pyyaml~=5.3.1",
        "tinydb~=3.15.2",
        "requests~=2.25.1",
        "pyjwt[crypto]~=2.0.1",
        "gql~=0.4.0",
        "semantic-version~=2.8.4",
        "kubernetes>=11.0.0",
        "retrying~=1.3.3",
    ],
    python_requires="~=3.7",
    packages=find_packages(),
    url="https://github.com/unikubehq/cli",
    project_urls={
        "Source": "https://github.com/unikubehq/cli",
        "Documentation": "https://cli.unikube.io",
        "Bug Tracker": "https://github.com/unikubehq/cli/issues",
    },
    author="Michael Schilonka",
    author_email="michael@blueshoe.de",
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points="""
        [console_scripts]
        unikube=unikube:cli
    """,
)



import re

import click
import click_spinner

import src.cli.console as console
from src import settings
from src.graphql import GraphQL
from src.keycloak.permissions import KeycloakPermissions
from src.local.providers.types import K8sProviderType
from src.storage.user import get_local_storage_user


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.pass_obj
def list(ctx, organization, **kwargs):
    """
    List all your projects.
    """

    context = ctx.context.get(organization=organization)

    # keycloak
    try:
        keycloak_permissions = KeycloakPermissions(authentication=ctx.auth)
        permission_list = keycloak_permissions.get_permissions_by_scope("project:*")
    except Exception as e:
        permission_list = None
        console.debug(e)
        console.exit_generic_error()

    # append "(active)"
    if context.project_id:
        for permission in permission_list:
            if permission.rsid == context.project_id:
                permission.rsid += " (active)"

    # console
    project_list = []
    for permission in permission_list:
        # parse name
        try:
            name = re.findall(r"project (.+?)\({rsid}\)".format(rsid=permission.rsid), permission.rsname)[0]
        except Exception:
            name = permission.rsname

        project_list.append(
            {
                "id": permission.rsid,
                "name": name,
            }
        )

    if not project_list:
        console.info("No projects available. Please go to https://app.unikube.io and create a project.")
        exit(0)

    console.table(
        data=project_list,
        headers={
            "id": "Id",
            "slug": "Identifier",
            "description": "Description",
        },
    )


@click.command()
@click.argument("project_title", required=False)
@click.pass_obj
def info(ctx, project_title, **kwargs):
    """
    Display further information of the selected project.
    """

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            {
                allProjects {
                    results {
                        id
                        title
                        description
                        specRepository
                        specRepositoryBranch
                    }
                }
            }
            """
        )
    except Exception:
        data = None
        console.exit_generic_error()

    project_list = data["allProjects"]["results"]

    # argument
    if not project_title:
        # argument from context
        context = ctx.context.get()
        if context.project_id:
            project = ctx.context.get_project()
            project_title = project["title"]

        # argument from console
        else:
            project_title = console.list(
                message="Please select a project",
                choices=[project["title"] for project in project_list],
            )
            if project_title is None:
                return None

    # select
    project_selected = None
    for project in project_list:
        if project["title"] == project_title:
            project_selected = project
            break

    # console
    if project_selected:
        project_selected["repository"] = project_selected.pop("specRepository")
        project_selected["repository branch"] = project_selected.pop("specRepositoryBranch")

        console.table(
            data={
                "key": [k for k in project_selected.keys()],
                "value": [v for v in project_selected.values()],
            },
            headers=["Key", "Value"],
        )
    else:
        console.error("Project does not exist.")


@click.command()
@click.argument("project_id", required=False)
@click.option("--remove", "-r", is_flag=True, default=False, help="Remove local organization context")
@click.pass_obj
def use(ctx, project_id, remove, **kwargs):
    """
    Set local project context.
    """

    # user_data / context
    local_storage_user = get_local_storage_user()
    user_data = local_storage_user.get()
    context = user_data.context

    # option: --remove
    if remove:
        user_data.context.deck_id = None
        user_data.context.project_id = None
        local_storage_user.set(user_data)
        console.success("Project context removed.")
        return None

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($organization_id: UUID) {
                allProjects(organizationId: $organization_id) {
                    results {
                        title
                        id
                        organization {
                            id
                        }
                    }
                }
            }
            """,
            query_variables={
                "organization_id": context.organization_id,
            },
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    project_list = data["allProjects"]["results"]
    project_dict = {project["id"]: project["title"] for project in project_list}

    # argument
    if not project_id:
        project_title = console.list(
            message="Please select a project",
            choices=project_dict.values(),
        )
        if project_title is None:
            return False

        for id, title in project_dict.items():
            if title == project_title:
                project_id = id

    project = project_dict.get(project_id, None)
    if not project:
        console.error(f"Unknown project with id: {project_id}.")

    # set project
    user_data.context.deck_id = None
    user_data.context.project_id = project["id"]
    user_data.context.organization_id = project["organization"]["id"]
    local_storage_user.set(user_data)

    console.success(f"Project context: {user_data.context}")


@click.command()
@click.argument("project_title", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--ingress", help="Specify the ingress port for the project", default=None)
@click.option(
    "--provider",
    "-p",
    help="Specify the Kubernetes provider type for this cluster",
    default=settings.UNIKUBE_DEFAULT_PROVIDER_TYPE.name,
)
@click.option("--workers", help="Specify count of k3d worker nodes", default=1)
@click.pass_obj
def up(ctx, project_title, organization, ingress, provider, workers, **kwargs):
    """
    Start/Resume a project cluster.
    """

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            {
                allProjects {
                    results {
                        id
                        title
                        clusterSettings {
                            id
                            port
                        }
                    }
                }
            }
            """
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    project_list = data["allProjects"]["results"]

    # argument
    if not project_title:
        # argument from context
        context = ctx.context.get(organization=organization)
        if context.project_id:
            project = ctx.context.get_project()
            project_title = project["title"]

        # argument from console
        else:
            cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)
            cluster_title_list = [item.name for item in cluster_list]

            project_list_choices = [item["title"] for item in project_list if item["title"] not in cluster_title_list]

            project_title = console.list(
                message="Please select a project",
                choices=project_list_choices,
            )
            if project_title is None:
                return False

    # check access to the project
    project_title_list = [project["title"] for project in project_list]
    if project_title not in project_title_list:
        console.info(f"The project '{project_title}' could not be found.")
        return None

    # get project id
    project_id = None
    for project in project_list:
        if project["title"] == project_title:
            project_id = project["id"]
            if ingress is None:
                ingress = project["clusterSettings"]["port"]
            break

    # cluster up
    cluster_data = ctx.cluster_manager.get(id=project_id)
    cluster_data.name = project_title
    ctx.cluster_manager.set(id=project_id, data=cluster_data)

    try:
        cluster_provider_type = K8sProviderType[provider]
    except KeyError:
        console.error(
            f"The provider '{provider}' is not supported. Please use "
            f"one of: {','.join(opt.name for opt in K8sProviderType)}",
            _exit=True,
        )

    cluster = ctx.cluster_manager.select(cluster_data=cluster_data, cluster_provider_type=cluster_provider_type)
    console.info(
        f"Setting up a Kubernetes cluster (with provider {provider}) for " f"project '{cluster.display_name}'."
    )

    if not cluster.exists():
        console.info(f"Kubernetes cluster for '{cluster.display_name}' does not exist, creating it now.")
        success = cluster.create(
            ingress_port=ingress,
            workers=workers,
        )

    # start
    else:
        if cluster.ready():
            console.info(f"Kubernetes cluster for '{cluster.display_name}' is already running.")
            return None

        else:
            console.info(f"Kubernetes cluster for '{cluster.display_name}' already exists, starting it now.")
            with click_spinner.spinner():
                success = cluster.start()

    # console
    if success:
        console.success("The project cluster is up.")
    else:
        console.error("The project cluster could not be started.")


@click.command()
@click.argument("project_title", required=False)
@click.pass_obj
def down(ctx, project_title, **kwargs):
    """
    Stop/Pause cluster.py
    """

    cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)
    cluster_title_list = [item.name for item in cluster_list]

    # argument
    if not project_title:
        # argument from context
        context = ctx.context.get()
        if context.project_id:
            project = ctx.context.get_project()
            project_title = project["title"]

        # argument from console
        else:
            project_title = console.list(
                message="Please select a project",
                message_no_choices="No cluster is running.",
                choices=cluster_title_list,
            )
            if project_title is None:
                return None

    # check if project is in local storage
    if project_title not in cluster_title_list:
        console.info("The project cluster could not be found.")
        return None

    # get cluster
    cluster = None
    for cluster_data in cluster_list:
        if cluster_data.name == project_title:
            cluster = ctx.cluster_manager.select(
                cluster_data=cluster_data,
            )
            break

    # cluster down
    if not cluster.exists():
        # something went wrong or cluster was already delete from somewhere else
        console.info(f"No Kubernetes cluster to stop for '{cluster.display_name}'")
        return None

    if not cluster.ready():
        console.info(f"Kubernetes cluster for '{cluster.display_name}' is not running")
        return None

    # stop cluster
    console.info(f"Stopping Kubernetes cluster for '{cluster.display_name}'")
    success = cluster.stop()

    # console
    if success:
        console.success("The project cluster is down.")
    else:
        console.error("The cluster could not be stopped.")


@click.command()
@click.argument("project_title", required=False)
@click.pass_obj
def delete(ctx, project_title, **kwargs):
    """Delete the current project and all related data"""

    cluster_list = ctx.cluster_manager.get_cluster_list()
    cluster_title_list = [item.name for item in cluster_list]

    # argument
    if not project_title:
        # argument from context
        context = ctx.context.get()
        if context.project_id:
            project = ctx.context.get_project()
            project_title = project["title"]

        # argument from console
        else:
            project_title = console.list(
                message="Please select a project",
                message_no_choices="No cluster available.",
                choices=cluster_title_list,
            )
            if project_title is None:
                return None

    # initial warning
    console.warning("Deleting a project will remove the custer including all of its data.")

    # confirm question
    confirm = input("Do want to continue [N/y]: ")
    if confirm not in ["y", "Y", "yes", "Yes"]:
        console.info("No action taken.")
        return None

    # get cluster
    cluster = None
    for cluster_data in cluster_list:
        if cluster_data.name == project_title:
            cluster = ctx.cluster_manager.select(
                cluster_data=cluster_data,
            )
            break

    # delete cluster
    if not cluster.exists():
        console.info(f"No Kubernetes cluster to delete for '{cluster.display_name}', nothing to do.")
        ctx.cluster_manager.delete(cluster.id)
        return None

    success = cluster.delete()

    # console
    if success:
        console.success("The project was deleted successfully.")
        ctx.cluster_manager.delete(cluster.id)
    else:
        console.error("The cluster could not be deleted.")



from getpass import getpass

import click
import jwt

import src.cli.console as console
from src import settings


@click.command()
@click.option("--email", "-e", type=str, help="Authentication email")
@click.option("--password", "-p", type=str, help="Authentication password")
@click.pass_obj
def login(ctx, email, password, **kwargs):
    """
    Authenticate with a Unikube host.
    """

    # email
    if not email:
        email = click.prompt("email", type=str)

    # password
    if not password:
        password = getpass("password:")

    response = ctx.auth.login(
        email,
        password,
    )
    if response["success"]:
        try:
            token = jwt.decode(
                response["response"]["access_token"],
                algorithms=["RS256"],
                audience=settings.TOKEN_AUDIENCE,
                options={"verify_signature": False},
            )
        except Exception as e:
            console.debug(e)
            console.debug(response)
            console.error("Login failed. Your token does not match.")
            return False

        if token["given_name"]:
            console.success(f'Login success. Hello {token["given_name"]}!')
        else:
            console.success("Login success.")

    else:
        console.error("Login failed. Please check your e-mail or re-enter your password.")

    return True


@click.command()
@click.pass_obj
def logout(ctx, **kwargs):
    """
    Log out of a Unikube host.
    """

    ctx.auth.logout()
    console.info("Logout completed.")

    return True


@click.command()
@click.pass_obj
def status(ctx, **kwargs):
    """
    View authentication status.
    """

    response = ctx.auth.verify()
    if response["success"]:
        console.success("Authentication verified.")
    else:
        console.info("Authentication could not be verified.")

    return True
